/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// console.c

#include "client.h"

console_t	con;

cvar_t		*con_notifytime;
cvar_t		*con_alpha;		// Knightare- Psychospaz's transparent console
cvar_t		*con_height;	// Knightmare- how far the console drops down


#define		MAXCMDLINE	256

#define CON_FONT_SCALE con_font_size->value/8

extern	char	key_lines[32][MAXCMDLINE];
extern	int		edit_line;
extern	int		key_linepos;
		

int stringLengthExtra (const char *string);
int stringLen (const char *string)
{
	return strlen(string) - stringLengthExtra(string);
}


// Knightmare- this sets the actual text color, can be called from anywhere
void TextColor (int colornum, int *red, int *green, int *blue)
{
	switch (colornum)
	{
		case 1:		//red
			*red =	255;
			*green=	0;
			*blue =	0;
			break;
		case 2:		//green
			*red =	0;
			*green=	255;
			*blue =	0;
			break;
		case 3:		//yellow
			*red =	255;
			*green=	255;
			*blue =	0;
			break;
		case 4:		//blue
			*red =	0;
			*green=	0;
			*blue =	255;
			break;
		case 5:		//teal
			*red =	0;
			*green=	255;
			*blue =	255;
			break;
		case 6:		//pink
			*red =	255;
			*green=	0;
			*blue =	255;
			break;
		case 7:		//white
			*red =	255;
			*green=	255;
			*blue =	255;
			break;
		case 8:		//black
			*red =	0;
			*green=	0;
			*blue =	0;
			break;
		case 9:		//orange
			*red =	255;
			*green=	135;
			*blue =	0;
			break;
		case 0:		//gray
			*red =	155;
			*green=	155;
			*blue =	155;
			break;
		default:	//white
			*red =	255;
			*green=	255;
			*blue =	255;
			break;
	}
}

qboolean setParams(char modifier, int *red, int *green, int *blue, int *bold, int *shadow, int *italic, int *reset, int *newline)
{
	switch (modifier)
	{
		case 'N':
		case 'n':
			*newline = true;
			return true;
		case 'R':
		case 'r':
			*reset = true;
			return true;
		case 'B':
		case 'b':
			if (*bold) 
				*bold = false;
			else 
				*bold=true;
			return true;
		case 'S':
		case 's':
			if (*shadow) 
				*shadow=false; 
			else 
				*shadow=true;
			return true;
		case 'I':
		case 'i':
			if (*italic) 
				*italic=false; 
			else 
				*italic=true;
			return true;
		case '1':	//red
		case '2':	//green
		case '3':	//yellow
		case '4':	//blue
		case '5':	//teal
		case '6':	//pink
		case '7':	//white
		case '8':	//black
		case '9':	//orange
		case '0':	//gray
			TextColor(atoi(&modifier), red, green, blue);
			return true;
		case 'A':	//alt text color
		case 'a':
			TextColor((int)alt_text_color->value, red, green, blue);
			return true;
	}
	
	return false;
}


#if 0
void DrawString (int x, int y, char *s, int alpha)
{
	while (*s)
	{
		R_DrawChar (x, y, *s, alpha);
		x+=FONT_SIZE;
		s++;
	}
}
#else
void DrawString (int x, int y, char *string, int alpha)
{
	unsigned i, j;
	char modifier, character;
	int len, red, green, blue, italic, shadow, bold, reset, newline;
	qboolean modified;

	//default
	red = 255;
	green = 255;
	blue = 255;
	italic = false;
	shadow = false;
	bold = false;

	len = strlen( string );
	for ( i = 0, j = 0; i < len; i++ )
	{
		modifier = string[i];
		if (modifier&128) modifier &= ~128;

		if (modifier == '^' && i < len)
		{
			i++;

			reset = 0;
			modifier = string[i];
			if (modifier&128) modifier &= ~128;

			if (modifier!='^')
			{
				modified = setParams(modifier, &red, &green, &blue, &bold, &shadow, &italic, &reset, &newline);

				if (reset)
				{
					red = 255;
					green = 255;
					blue = 255;
					italic = false;
					shadow = false;
					bold = false;
				}
				if (modified)
					continue;
				else
					i--;
			}
		}
		j++;


		character = string[i];
		if (bold && character<128)
			character += 128;
		else if (bold && character>128)
			character -= 128;

		// Knightmare- hack for alternate text color
		if (!modified && (character & 128))
			//red = blue = 0;
			TextColor((int)alt_text_color->value, &red, &green, &blue);

		if (shadow)
			R_DrawScaledChar( ( x + (j-1)*FONT_SIZE+FONT_SIZE/4 ), y+(FONT_SIZE/8),  character, CON_FONT_SCALE, 0, 0, 0, alpha, italic);
		R_DrawScaledChar( ( x + (j-1)*FONT_SIZE ), y, character, CON_FONT_SCALE, red, green, blue, alpha, italic);
	}
}
#endif


#if 0
void DrawAltString (int x, int y, char *s, int alpha)
{
	while (*s)
	{
		R_DrawChar (x, y, *s ^ 0x80, alpha);
		x+=FONT_SIZE;
		s++;
	}
}
#else
void DrawAltString (int x, int y, char *string, int alpha)
{
	unsigned i, j;
	char modifier, character;
	int len, red, green, blue, italic, shadow, bold, reset, newline;
	qboolean modified;

	//default
	red = 255;
	green = 255;
	blue = 255;
	italic = false;
	shadow = false;
	bold = true;
	

	len = strlen( string );
	for ( i = 0, j = 0; i < len; i++ )
	{
		modifier = string[i];
		if (modifier&128) modifier &= ~128;

		if (modifier == '^' && i < len)
		{
			i++;

			reset = 0;
			modifier = string[i];
			if (modifier&128) modifier &= ~128;

			if (modifier!='^')
			{
				modified = setParams(modifier, &red, &green, &blue, &bold, &shadow, &italic, &reset, &newline);

				if (reset)
				{
					red = 255;
					green = 255;
					blue = 255;
					italic = false;
					shadow = false;
					bold = true;
				}
				if (modified)
					continue;
				else
					i--;
			}
		}
		j++;

		character = string[i];
		if (bold && character<128)
			character += 128;
		else if (bold && character>128)
			character -= 128;

		// Knightmare- hack for alternate text color
		if (!modified) {
			character ^= 0x80;
			if (character & 128)
				//red = blue = 0;
				TextColor((int)alt_text_color->value, &red, &green, &blue);
		}

		if (shadow)
			R_DrawScaledChar( ( x + (j-1)*FONT_SIZE+FONT_SIZE/4 ), y+(FONT_SIZE/8),  character,1.0f, 0, 0, 0, alpha, italic);
		R_DrawScaledChar( ( x + (j-1)*FONT_SIZE ), y, character ,1.0f, red, green, blue, alpha, italic);
	}
}
#endif


void Key_ClearTyping (void)
{
	key_lines[edit_line][1] = 0;	// clear any typing
	key_linepos = 1;
	con.backedit = 0;
}

/*
================
Con_ToggleConsole_f
================
*/
void Con_ToggleConsole_f (void)
{
//	SCR_EndLoadingPlaque ();	// get rid of loading plaque

	// Knightmare- allow disconnected menu
	if (cls.state == ca_disconnected && cls.key_dest != key_menu)
	{	// start the demo loop again
		Cbuf_AddText ("d1\n");
		return;
	}

	/*if (cl.attractloop)
	{
		Cbuf_AddText ("killserver\n");
		return;
	}*/

	Key_ClearTyping ();
	Con_ClearNotify ();

	if (cls.consoleActive) // Knightmare changed
	//if (cls.key_dest == key_console)
	{
		cls.consoleActive = false; // Knightmare added

		//M_ForceMenuOff ();
		if (cls.key_dest != key_menu)
			Cvar_Set ("paused", "0");
	}
	else
	{
		cls.consoleActive = true; // Knightmare added
		//M_ForceMenuOff ();
		//cls.key_dest = key_console;	

		if (Cvar_VariableValue ("maxclients") == 1 
			&& Com_ServerState () && cls.key_dest != key_menu)
			Cvar_Set ("paused", "1");
	}
}

/*
================
Con_ToggleChat_f
================
*/
void Con_ToggleChat_f (void)
{
	Key_ClearTyping ();

	if (cls.consoleActive) // Knightmare added
	//if (cls.key_dest == key_console)
	{
		if (cls.state == ca_active)
		{
			M_ForceMenuOff ();
			cls.consoleActive = false; // Knightmare added
			cls.key_dest = key_game;
		}
	}
	else
		//cls.key_dest = key_console;
		cls.consoleActive = true; // Knightmare added

	Con_ClearNotify ();
}

/*
================
Con_Clear_f
================
*/
void Con_Clear_f (void)
{
	memset (con.text, ' ', CON_TEXTSIZE);
}

						
/*
================
Con_Dump_f

Save the console contents out to a file
================
*/
void Con_Dump_f (void)
{
	int		l, x;
	char	*line;
	FILE	*f;
	char	buffer[1024];
	char	name[MAX_OSPATH];

	if (Cmd_Argc() != 2)
	{
		Com_Printf ("usage: condump <filename>\n");
		return;
	}

	Com_sprintf (name, sizeof(name), "%s/%s.txt", FS_Gamedir(), Cmd_Argv(1));

	Com_Printf ("Dumped console text to %s.\n", name);
	FS_CreatePath (name);
	f = fopen (name, "w");
	if (!f)
	{
		Com_Printf ("ERROR: couldn't open.\n");
		return;
	}

	// skip empty lines
	for (l = con.current - con.totallines + 1 ; l <= con.current ; l++)
	{
		line = con.text + (l%con.totallines)*con.linewidth;
		for (x=0 ; x<con.linewidth ; x++)
			if (line[x] != ' ')
				break;
		if (x != con.linewidth)
			break;
	}

	// write the remaining lines
	buffer[con.linewidth] = 0;
	for ( ; l <= con.current ; l++)
	{
		line = con.text + (l%con.totallines)*con.linewidth;
		strncpy (buffer, line, con.linewidth);
		for (x=con.linewidth-1 ; x>=0 ; x--)
		{
			if (buffer[x] == ' ')
				buffer[x] = 0;
			else
				break;
		}
		for (x=0; buffer[x]; x++)
			buffer[x] &= 0x7f;

		fprintf (f, "%s\n", buffer);
	}

	fclose (f);
}

						
/*
================
Con_ClearNotify
================
*/
void Con_ClearNotify (void)
{
	int		i;
	
	for (i=0 ; i<NUM_CON_TIMES ; i++)
		con.times[i] = 0;
}

						
/*
================
Con_MessageMode_f
================
*/
void Con_MessageMode_f (void)
{
	chat_team = false;
	cls.key_dest = key_message;
	cls.consoleActive = false; // Knightmare added
}

/*
================
Con_MessageMode2_f
================
*/
void Con_MessageMode2_f (void)
{
	chat_team = true;
	cls.key_dest = key_message;
	cls.consoleActive = false; // Knightmare added
}

/*
================
Con_CheckResize

If the line width has changed, reformat the buffer.
================
*/
void Con_CheckResize (void)
{
	int		i, j, width, oldwidth, oldtotallines, numlines, numchars;
	char	tbuf[CON_TEXTSIZE];

	if (con_font_size)
		width = viddef.width/con_font_size->value - 2;
	else	 // (viddef.width / 8)
		width = (viddef.width >> 3) - 2;

	if (width == con.linewidth)
		return;

	if (width < 1)		// video hasn't been initialized yet
	{
		width = 38;
		con.linewidth = width;
		con.backedit = 0;
		con.totallines = CON_TEXTSIZE / con.linewidth;
		memset (con.text, ' ', CON_TEXTSIZE);
	}
	else
	{
		oldwidth = con.linewidth;
		con.linewidth = width;
		con.backedit = 0;
		oldtotallines = con.totallines;
		con.totallines = CON_TEXTSIZE / con.linewidth;
		numlines = oldtotallines;

		if (con.totallines < numlines)
			numlines = con.totallines;

		numchars = oldwidth;
	
		if (con.linewidth < numchars)
			numchars = con.linewidth;

		memcpy (tbuf, con.text, CON_TEXTSIZE);
		memset (con.text, ' ', CON_TEXTSIZE);

		for (i=0 ; i<numlines ; i++)
		{
			for (j=0 ; j<numchars ; j++)
			{
				con.text[(con.totallines - 1 - i) * con.linewidth + j] =
						tbuf[((con.current - i + oldtotallines) %
							  oldtotallines) * oldwidth + j];
			}
		}

		Con_ClearNotify ();
	}

	con.current = con.totallines - 1;
	con.display = con.current;
}


/*
================
Con_Init
================
*/
void Con_Init (void)
{
	con.linewidth = -1;
	con.backedit = 0;

	Con_CheckResize ();
	
	Com_Printf ("Console initialized.\n");

//
// register our commands
//
	con_notifytime = Cvar_Get ("con_notifytime", "4", 0); // Knightmare- increased for fade
	// Knightmare- Psychospaz's transparent console
	con_alpha = Cvar_Get ("con_alpha", "0.75", CVAR_ARCHIVE);
	// Knightmare- how far the console drops down
	con_height = Cvar_Get ("con_height", "0.5", CVAR_ARCHIVE);

	Cmd_AddCommand ("toggleconsole", Con_ToggleConsole_f);
	Cmd_AddCommand ("togglechat", Con_ToggleChat_f);
	Cmd_AddCommand ("messagemode", Con_MessageMode_f);
	Cmd_AddCommand ("messagemode2", Con_MessageMode2_f);
	Cmd_AddCommand ("clear", Con_Clear_f);
	Cmd_AddCommand ("condump", Con_Dump_f);
	con.initialized = true;
}


/*
===============
Con_Linefeed
===============
*/
void Con_Linefeed (void)
{
	con.x = 0;
	if (con.display == con.current)
		con.display++;
	con.current++;
	memset (&con.text[(con.current%con.totallines)*con.linewidth]
	, ' ', con.linewidth);
}

/*
================
Con_Print

Handles cursor positioning, line wrapping, etc
All console printing must go through this in order to be logged to disk
If no console is visible, the text will appear at the top of the game window
================
*/
void Con_Print (char *txt)
{
	int		y;
	int		c, l;
	static int	cr;
	int		mask;

	if (!con.initialized)
		return;

	if (txt[0] == 1 || txt[0] == 2)
	{
		mask = 128;		// go to colored text
		txt++;
	}
	else
		mask = 0;


	while ( (c = *txt) )
	{
	// count word length
		for (l=0 ; l< con.linewidth ; l++)
			if ( txt[l] <= ' ')
				break;

	// word wrap
		if (l != con.linewidth && (con.x + l > con.linewidth) )
			con.x = 0;

		txt++;

		if (cr)
		{
			con.current--;
			cr = false;
		}

		
		if (!con.x)
		{
			Con_Linefeed ();
		// mark time for transparent overlay
			if (con.current >= 0)
				con.times[con.current % NUM_CON_TIMES] = cls.realtime;
		}

		switch (c)
		{
		case '\n':
			con.x = 0;
			break;

		case '\r':
			con.x = 0;
			cr = 1;
			break;

		default:	// display character and advance
			y = con.current % con.totallines;
			con.text[y*con.linewidth+con.x] = c | mask | con.ormask;
			con.x++;
			if (con.x >= con.linewidth)
				con.x = 0;
			break;
		}
		
	}
}


/*
==============
Con_CenteredPrint
==============
*/
void Con_CenteredPrint (char *text)
{
	int		l;
	char	buffer[1024];

	l = strlen(text);
	l = (con.linewidth-l)/2;
	if (l < 0)
		l = 0;
	memset (buffer, ' ', l);
	strcpy (buffer+l, text);
	strcat (buffer, "\n");
	Con_Print (buffer);
}

/*
==============================================================================

DRAWING

==============================================================================
*/


/*
================

Con_DrawInput
The input line scrolls horizontally if typing goes beyond the right edge
================
*/
void Con_DrawInput (void)
{
	int		y;
	int		i;
	char	*text, output[2048];

	if (!cls.consoleActive && cls.state == ca_active)
		return;		// don't draw anything (always draw if not active)

	/*if (cls.key_dest == key_menu)
		return;
	if (cls.key_dest != key_console && cls.state == ca_active)
		return;*/		// don't draw anything (always draw if not active)

	text = key_lines[edit_line];
	
// add the cursor frame
	if (con.backedit)
		text[key_linepos] = ' ';
	else
		text[key_linepos] = 10+((int)(cls.realtime>>8)&1);
	
// fill out remainder with spaces
	for (i=key_linepos+1 ; i< con.linewidth ; i++)
		text[i] = ' ';
		
//	prestep if horizontally scrolling
	if (key_linepos >= con.linewidth)
		text += 1 + key_linepos - con.linewidth;
		
// draw it
	y = con.vislines-FONT_SIZE*2; // was 16

	Com_sprintf (output, sizeof(output), "");
	for (i=0 ; i<con.linewidth ; i++)
	{
		if (con.backedit == key_linepos-i && ((int)(cls.realtime>>8)&1))
			Com_sprintf (output, sizeof(output), "%s%c", output, 11 );
		else
			Com_sprintf (output, sizeof(output), "%s%c", output, text[i]);
	}
	DrawString ( FONT_SIZE/2, con.vislines - (int)(2.75*FONT_SIZE), output, 255);
	//for (i=0; i<con.linewidth; i++)
	//	R_DrawChar ( (i+1)*FONT_SIZE, con.vislines - (int)(2.75*FONT_SIZE), text[i], 255);

// remove cursor
	key_lines[edit_line][key_linepos] = 0;
}


/*
================
Con_DrawNotify

Draws the last few lines of output transparently over the game top
================
*/
void Con_DrawNotify (void)
{
	int		x;
	char	*text, output[2048];
	int		i, j;
	//int		time;
	char	*s;
	int		alpha, lines;
	float	v, time;

	lines = 0;

	v = 0;

	Com_sprintf (output, sizeof(output), "");

	//this is the say msg while typeing...
	if (cls.key_dest == key_message)
	{
		if (chat_team)
			Com_sprintf (output, sizeof(output), "%s", " say_team: ");
		else
			Com_sprintf (output, sizeof(output), "%s", " say: ");

		s = chat_buffer;
		x = 0;
		if (chat_bufferlen > (viddef.width/FONT_SIZE)-(strlen(output)+1))
			x += chat_bufferlen - (int)((viddef.width/FONT_SIZE)-(strlen(output)+1));

		while(s[x])
		{
			if (chat_backedit && chat_backedit == chat_bufferlen-x && ((int)(cls.realtime>>8)&1))
				Com_sprintf (output, sizeof(output), "%s%c", output, 11 );
			else
				Com_sprintf (output, sizeof(output), "%s%c", output, (char)s[x]);

			x++;
		}

		if (!chat_backedit)
			Com_sprintf (output, sizeof(output), "%s%c", output, 10+((int)(cls.realtime>>8)&1) );		

		DrawString (0, v, output, 255);

		v += FONT_SIZE*2; //make extra space so we have room
	}

	for (i= con.current-NUM_CON_TIMES+1; i<=con.current; i++)
	{
		if (i < 0)
			continue;
		time = con.times[i % NUM_CON_TIMES];
		if (time == 0)
			continue;
		time = cls.realtime - time;
		if (time > con_notifytime->value*1000)
			continue;
		//text = con.text + (i % con.totallines)*con.linewidth;
		//for (x = 0 ; x < con.linewidth ; x++)
		//	R_DrawChar ( (x+1)<<3, v, text[x], 255);
		//v += 8;

		//vertical offset set by closest to dissapearing
		lines++;
	}

	//v = 0;
	if (lines)
		for (j=0, i= con.current-NUM_CON_TIMES+1 ; i<=con.current ; i++, j++)
		{
			if (i < 0)
				continue;
			time = con.times[i % NUM_CON_TIMES];
			if (time == 0)
				continue;
			time = cls.realtime - time;
			if (time > con_notifytime->value*1000)
				continue;

			text = con.text + (i % con.totallines)*con.linewidth;
			
			alpha = 255 * sqrt( (1.0-time/(con_notifytime->value*1000.0+1.0)) * (((float)v+8.0)) / (8.0*lines) );
			//alpha = (1-time/(con_notifytime->value*1000.0+2)) * (255*(v+16)) / (4*lines);
			if (alpha < 0) alpha=0;
			if (alpha > 255) alpha=255;

			Com_sprintf (output, sizeof(output), "");
			for (x = 0 ; x < con.linewidth ; x++)
				Com_sprintf (output, sizeof(output), "%s%c", output, (char)text[x]);
				//R_DrawChar ( (x+1)*FONT_SIZE, v, text[x], alpha);

			DrawString (FONT_SIZE/2, v, output, alpha);

			v += FONT_SIZE;
		}

	//this is the say msg while typeing...
	/*if (cls.key_dest == key_message)
	{
		if (chat_team)
		{
			DrawString (8, v, "say_team:", 255);
			skip = 11;
		}
		else
		{
			DrawString (8, v, "say:", 255);
			skip = 5;
		}

		s = chat_buffer;
		if (chat_bufferlen > (viddef.width/FONT_SIZE)-(skip+1))
			s += chat_bufferlen - ((int)(viddef.width/FONT_SIZE)-(skip+1));
		//if (chat_bufferlen > (viddef.width>>3)-(skip+1))
		//	s += chat_bufferlen - ((viddef.width>>3)-(skip+1));
		x = 0;
		while(s[x])
		{
			R_DrawChar ( (x+skip)*FONT_SIZE, v, s[x], 255);
			//R_DrawChar ( (x+skip)<<3, v, s[x], 255);
			x++;
		}
		R_DrawChar ( (x+skip)*FONT_SIZE, v, 10+((cls.realtime>>8)&1), 255);
		//R_DrawChar ( (x+skip)<<3, v, 10+((cls.realtime>>8)&1), 255);
		v += FONT_SIZE; //was 8
	}*/
	
	if (v)
	{
		SCR_AddDirtyPoint (0,0);
		SCR_AddDirtyPoint (viddef.width-1, v);
	}
}

/*
================
Con_DrawConsole

Draws the console with the solid background
================
*/

//void Con_DrawConsole (float frac)
void Con_DrawConsole (float frac, qboolean ingame) //Knightmare
{
	int				i, j, x, y;//, n;
	int				rows;
	char			*text, output[1024];
	int				row;
	int				lines;
	char			version[64];
	char			dlbar[1024];
	float			alpha; //Knightmare added

	lines = viddef.height * frac;
	if (lines <= 0)
		return;

	if (lines > viddef.height)
		lines = viddef.height;

	//Knightare- Psychospaz's transparent console
	if (!ingame)
		alpha = 1;
	else
	//	alpha = (frac*con_alpha->value*2);
		alpha = ((frac/con_height->value)*con_alpha->value);
	//end Knightmare

// draw the background
	//Knightmare added alpha parm for Psychospaz's transparent console
	R_DrawStretchPic (0, -viddef.height+lines, viddef.width, viddef.height, "conback", alpha);
	SCR_AddDirtyPoint (0,0);
	SCR_AddDirtyPoint (viddef.width-1,lines-1);

	// Knightmare- changed to "KMQuake2 vx.xx"
#ifdef ERASER_COMPAT_BUILD
#ifdef NET_SERVER_BUILD
	Com_sprintf (version, sizeof(version), "^b^s^aKMQuake2 v%4.2f (Eraser net server)", VERSION);
#else // NET_SERVER_BUILD
	Com_sprintf (version, sizeof(version), "^b^s^aKMQuake2 v%4.2f (Eraser compatible)", VERSION);
#endif // NET_SERVER_BUILD
#else // ERASER_COMPAT_BUILD
#ifdef NET_SERVER_BUILD
	Com_sprintf (version, sizeof(version), "^b^s^aKMQuake2 v%4.2f (net server)", VERSION);
#else
	Com_sprintf (version, sizeof(version), "^b^s^aKMQuake2 v%4.2f", VERSION);
#endif // ERASER_COMPAT_BUILD
#endif // NEW_ENTITY_STATE_MEMBERS

	DrawString(viddef.width-FONT_SIZE*(stringLen((const char *)&version))-3, lines-FONT_SIZE-1, version, 255);
	//for (x  =0; x < 14; x++) //was x<5
	//	R_DrawChar (viddef.width-(14*FONT_SIZE+3)+x*FONT_SIZE, lines-(int)(FONT_SIZE*1.5), 128 + version[x], 255); //width- was -44, width = 115

// draw the text
	con.vislines = lines;
	
	rows = (lines-(int)(2.75*FONT_SIZE))/FONT_SIZE;		// rows of text to draw
	y = lines - (int)(3.75*FONT_SIZE);

// draw from the bottom up
	if (con.display != con.current)
	{
	// draw arrows to show the buffer is backscrolled
		for (x=0; x<con.linewidth; x+=(int)FONT_SIZE/2)
			R_DrawScaledChar( (x+1)*FONT_SIZE, y, '^', CON_FONT_SCALE, 255, 255, 255, 255, false);
			//R_DrawChar ( (x+1)*FONT_SIZE, y, '^', 255);
	
		y -= FONT_SIZE;
		rows--;
	}
	
	row = con.display;
	for (i=0; i<rows; i++, y-=FONT_SIZE, row--)
	{
		if (row < 0)
			break;
		if (con.current - row >= con.totallines)
			break;		// past scrollback wrap point
			
		text = con.text + (row % con.totallines)*con.linewidth;

		Com_sprintf (output, sizeof(output), "");
		for (x=0; x<con.linewidth; x++)
			Com_sprintf (output, sizeof(output), "%s%c", output, text[x]);
			//R_DrawChar ( (x+1)*FONT_SIZE, y, text[x], 255);
		DrawString ( 4, y, output, 256);
	}

//ZOID
	// draw the download bar
	// figure out width
	if (cls.download)
	{
		int graph_x, graph_y, graph_h, graph_w;
		// Knightmare- changeable download/map load bar color
		int red, green, blue;
		TextColor((int)alt_text_color->value, &red, &green, &blue);

		if ((text = strrchr(cls.downloadname, '/')) != NULL)
			text++;
		else
			text = cls.downloadname;

		memset (dlbar, 0, sizeof(dlbar)); // clear dlbar

		// make this a little shorter in case of longer version string
		x = con.linewidth - ((con.linewidth * 7) / 40) - (stringLen((const char *)&version)-14);

		y = x - strlen(text) - 8;
		i = con.linewidth/3;
		if (strlen(text) > i)
		{
			y = x - i - 11;
			strncpy(dlbar, text, i);
			dlbar[i] = 0;
			strcat(dlbar, "...");
		}
		else
			strcpy(dlbar, text);
		strcat(dlbar, ": ");
		i = strlen(dlbar);

		/*dlbar[i++] = '\x80';
		// where's the dot go?
		if (cls.downloadpercent == 0)
			n = 0;
		else
			n = y * cls.downloadpercent / 100;
			
		for (j = 0; j < y; j++)
			if (j == n)
				dlbar[i++] = '\x83';
			else
				dlbar[i++] = '\x81';
		dlbar[i++] = '\x82';
		dlbar[i] = 0;*/
		
		// init solid color download bar
		graph_x = (i+1)*FONT_SIZE;
		graph_y = con.vislines - (int)(FONT_SIZE*1.5); // was -12
		graph_w = y*FONT_SIZE;
		graph_h = FONT_SIZE;

		for (j = 0; j < y; j++) // add blank spaces
			sprintf(dlbar + strlen(dlbar), " ");
		sprintf(dlbar + strlen(dlbar), " %2d%%", cls.downloadpercent);

		// draw it
		y = graph_y;
		for (i = 0; i < strlen(dlbar); i++)
			if (dlbar[i] != ' ')
				//R_DrawChar ( (i+1)*FONT_SIZE, y, dlbar[i], 255);
				R_DrawScaledChar( (i+1)*FONT_SIZE, y, dlbar[i], CON_FONT_SCALE, 255, 255, 255, 255, false);

		// new solid color download bar
		graph_y = y;
		R_DrawFill2 (
			graph_x-1, graph_y-1, graph_w+2, graph_h+2,
			255, 255, 255, 90); //was 127

		R_DrawFill2 (
			graph_x+1, graph_y+1,
			(int)((graph_w-2)*cls.downloadpercent*0.01), graph_h-2,
			red, green, blue, 255);
		// end Knightmare
	}
//ZOID

// draw the input prompt, user text, and cursor if desired
	Con_DrawInput ();
}


